/*
 * Partially implemented disassembler for Samsung SSP160x processors
 * (C) Copyright 2007-2008, Grazvydas "notaz" Ignotas
 */

#include "ssp.hpp"
#include <entry.hpp>
#include <srarea.hpp>



/*
 * Kernel event handler
 *
 * @param msgid Message ID to handle
 * @param ... Variable list of arguments
 * @return 1 on success
 */
static int notify(processor_t::idp_notify msgid, ...) {
  va_list va;
  va_start(va, msgid);

  // do IDA callbacks
  int code = invoke_callbacks(HT_IDP, msgid, va);
  if (code) return code;

  //msg("notify: %i  %i\n", msgid, processor_t::newfile);

  // deal with notification codes
  switch(msgid) {
  case processor_t::init:
    inf.mf = 1;       // Set big endian mode in the IDA kernel
    break;

  case processor_t::make_data: {
    ea_t ea = va_arg(va, ea_t);
    va_arg(va, ea_t); // skip
    va_arg(va, tid_t); // skip
    asize_t len = va_arg(va, asize_t);
    char buff[16];
    if (get_word(ea) < 0x400) break;
    qsnprintf(buff, sizeof(buff), "gloc_%04X", get_word(ea) << 1);
    if (len == 2) {
      set_cmt(ea, buff, 0);
      set_name(get_word(ea) << 1, buff);
    }
    else
    {
      set_cmt(ea, "", 0);
      set_name(get_word(ea) << 1, "");
    }
    //msg("notify: ea=%x flags=%x tid=%x len=%i\n", ea, flags, tid, len);
    break;
  }

  case processor_t::undefine: {
    ea_t ea = va_arg(va, ea_t);
    char buff[16], buff2[16] = { 0, };
    qsnprintf(buff, sizeof(buff), "gloc_%04X", get_word(ea) << 1);
    get_cmt(ea, 0, buff2, sizeof(buff2));
    if (strcmp(buff, buff2) == 0)
      // remove our auto comment
      set_cmt(ea, "", 0);
    break;
  }

  default:
    break;
  }
  va_end(va);

  // OK
  return(1);
}


/************************************************************************/
/* Register names                                                       */
/************************************************************************/
char *ssp_reg_names[32+8+8+2] = {
  "-",    "X",     "Y",    "A",
  "ST",   "STACK", "PC",   "P",

  "PM0",  "PM1",   "PM2",  "XST",
  "PM4",  "gr13",  "PMC",  "AL",

  "gr16", "gr17",  "gr18", "gr19",
  "gr20", "gr21",  "gr22", "gr23",
  "gr24", "gr25",  "gr26", "gr27",
  "gr28", "gr29",  "gr30", "gr31",
  // rij (ram pointer regs)?
  "r0",   "r1",    "r2",   "r3",
  "r4",   "r5",    "r6",   "r7",
  // erij (>= 1605 only?)?
  "er0",  "er1",   "er2",  "er3",
  "er4",  "er5",   "er6",  "er7",
  "cs",   "ds" // virtual regs required for IDA
/*
  "AL",   "?",     "?",    "XST",
  "PL",   "?",     "?",    "SRCR",
  "BRCR", "BRER",  "XRD0", "XRD1",
  "AE",   "DIOR",  "GR22", "GR23",
  "EXT0", "EXT1",  "EXT2", "EXT3",
  "EXT4", "EXT5",  "EXT6", "EXT7"
*/
};


static const char *ssp_headers[] = {
  "",
  "",
  NULL
};


static asm_t ssp_asm = {
  AS_COLON,
  0,
  "SSP1601 guesswork assembler",
  0,
  (const char**) ssp_headers,
  NULL,         // no bad instructions
  "org",
  "end",

  ";",          // comment string
  '\'',         // string delimiter
  '\'',         // char delimiter
  "+_*/%&|^()<>!+$@#.,\'\"?",    // special symbols in char+string constants

  "db",         // ascii string directive
  "db",         // byte directive
  "dw",         // word directive
  NULL,         // dword  (4 bytes)
  NULL,         // qword  (8 bytes)
  NULL,         // oword  (16 bytes)
  NULL,         // float  (4 bytes)
  NULL,         // double (8 bytes)
  NULL,         // tbyte  (10/12 bytes)
  NULL,         // packed decimal real
  "#d dup(#v)", // arrays (#h,#d,#v,#s(...)
  NULL,         // uninited arrays
  "equ",        // equ
  NULL,         // seg prefix
  NULL, NULL, NULL,
  NULL,
  "$",
  NULL,         // func_header
  NULL,         // func_footer
  NULL,         // public
  NULL,         // weak
  NULL,         // extrn
  NULL,         // comm
  NULL,         // get_type_name
  NULL,         // align
  '(', ')',     // lbrace, rbrace
  NULL,    // mod
  NULL,    // and
  NULL,    // or
  NULL,    // xor
  "~",     // not
  NULL,    // shl
  NULL,    // shr
  NULL,    // sizeof
  0,
};

/************************************************************************/
/* Assemblers supported by this module                                  */
/************************************************************************/
static asm_t *asms[] = { &ssp_asm, NULL };


/************************************************************************/
/* Short names of processor                                             */
/************************************************************************/
static char *shnames[] =
{
  "SSP1601",
  NULL
};

/************************************************************************/
/* Long names of processor                                              */
/************************************************************************/
static char *lnames[] = {
  "Samsung DSP: SSP1601",
  NULL
};



//--------------------------------------------------------------------------
// Opcodes of "return" instructions. This information will be used in 2 ways:
//      - if an instruction has the "return" opcode, its autogenerated label
//        will be "locret" rather than "loc".
//      - IDA will use the first "return" opcode to create empty subroutines.

//static uchar retcode_1[] = { 0xAF };
//static uchar retcode_2[] = { 0xBF };

static bytes_t retcodes[] = {
// { sizeof(retcode_1), retcode_1 },
// { sizeof(retcode_2), retcode_2 },
 { 0, NULL }                            // NULL terminated array
};



//-----------------------------------------------------------------------
//      Processor Definition
//-----------------------------------------------------------------------

extern "C" processor_t LPH =
{
  IDP_INTERFACE_VERSION,// version
  0x8123,               // id
  PR_RNAMESOK,          // can use register names for byte names
  8,                    // 8 bits in a byte for code segments
  8,                    // 8 bits in a byte for other segments

  shnames,              // array of short processor names
                        // the short names are used to specify the processor
                        // with the -p command line switch)
  lnames,               // array of long processor names
                        // the long names are used to build the processor type
                        // selection menu

  asms,                 // array of target assemblers

  notify,               // the kernel event notification callback

  header,               // generate the disassembly header
  footer,               // generate the disassembly footer

  segstart,             // generate a segment declaration (start of segment)
  std_gen_segm_footer,  // generate a segment footer (end of segment)

  NULL,                 // generate 'assume' directives

  ana,                  // analyze an instruction and fill the 'cmd' structure
  emu,                  // emulate an instruction

  out,                  // generate a text representation of an instruction
  outop,                // generate a text representation of an operand
  out_data,             // generate a text representation of a data item
  NULL,                 // compare operands
  NULL,                 // can an operand have a type?

  qnumber(ssp_reg_names), // Number of registers
  ssp_reg_names,        // Register names
  NULL,                 // get abstract register

  0,                    // Number of register files
  NULL,                 // Register file names
  NULL,                 // Register descriptions
  NULL,                 // Pointer to CPU registers

  rVcs,rVds,
  0,                    // size of a segment register
  rVcs,rVds,

  NULL,                 // No known code start sequences
  NULL, //retcodes,

  0,SSP_last,
  Instructions
};

